@@ -40,6 +40,21 @@ class TestOvsdbIdl(base.FunctionalTestCase):
         cleanup_cmd = self.api.del_br(self.brname)
         self.addCleanup(cleanup_cmd.execute)
 
+    def test_idl_run_exception_terminates(self):
+        run = self.api.idl.run
+        with mock.patch.object(self.api.idl, "run") as runmock:
+            exceptions = iter([Exception("TestException")])
+
+            def side_effect():
+                try:
+                    raise next(exceptions)
+                except StopIteration:
+                    return run()
+
+            runmock.side_effect = side_effect
+            exists = self.api.br_exists(self.brname).execute(check_error=True)
+            self.assertFalse(exists)
+
     def test_br_exists_false(self):
         exists = self.api.br_exists(self.brname).execute(check_error=True)
         self.assertFalse(exists)
@@ -104,11 +119,12 @@ class TestOvsdbIdl(base.FunctionalTestCase):
     def test_connection_reconnect(self):
         self.api.ovsdb_connection.stop()
         existsCmd = self.api.br_exists(self.brname)
-        self.assertRaises(exc.NotConnectedError,
-                          existsCmd.execute, check_error=True)
+        txn = self.api.create_transaction(check_error=True)
+        txn.add(existsCmd)
+        self.api.ovsdb_connection.queue_txn(txn)
         self.api.ovsdb_connection.start()
-        exists = self.api.br_exists(self.brname).execute(check_error=True)
-        self.assertFalse(exists)
+        result = txn.results.get(timeout=self.api.ovsdb_connection.timeout)
+        self.assertEqual(result, [False])
 
     def test_connection_disconnect_timeout(self):
         _is_running_mock = mock.PropertyMock(return_value=True)
